==> pdstools-3.4.3.tar.gz: pdstools-3.4.3/python/pdstools/prediction/Prediction.py <==
from collections import namedtuple
from typing import List
import polars as pl

from ..utils import cdh_utils


class Prediction:
    predictions = None

    # These are pretty strict conditions - many configurations appear not to satisfy these
    # perhaps the Total = Test + Control is no longer met when Impact Analyzer is around
    validityExpr = (
        (pl.col("Positives") > 0)
        & (pl.col("Positives_Test") > 0)
        & (pl.col("Positives_Control") > 0)
        & (pl.col("Negatives") > 0)
        & (pl.col("Negatives_Test") > 0)
        & (pl.col("Negatives_Control") > 0)
        # & (
        #     pl.col("Positives")
        #     == (pl.col("Positives_Test") + pl.col("Positives_Control"))
        # )
        # & (
        #     pl.col("Negatives")
        #     == (pl.col("Negatives_Test") + pl.col("Negatives_Control"))
        # )
    )

    # Prediction name (case insensitive), channel, direction, is-standard (True here), is-multi channel
    NBAD_predictions = namedtuple( 
        "Configurations",
        ["model_name", "channel", "direction", "standard", "multi_channel"],
    ) 
    standardNBADPredictionsList:List[NBAD_predictions] = [
        ("PredictWebPropensity", "Web", "Inbound", True, False),
        ("PredictMobilePropensity", "Mobile", "Inbound", True, False),
        ("PredictOutboundEmailPropensity", "E-mail", "Outbound", True, False),
        ("PredictOutboundPushPropensity", "Push", "Outbound", True, False),
        ("PredictOutboundSMSPropensity", "SMS", "Outbound", True, False),
        ("PredictInboundRetailPropensity", "Retail", "Inbound", True, False),
        ("PredictOutboundRetailPropensity", "Retail", "Outbound", True, False),
        ("PredictInboundCallCenterPropensity", "Call Center", "Inbound", True, False),
        ("PredictOutboundCallCenterPropensity", "Call Center", "Outbound", True, False),
        ("PredictInboundDefaultPropensity", "Default", "Inbound", True, False),
        ("PredictOutboundDefaultPropensity", "Default", "Outbound", True, False),
        ("PredictInboundOtherPropensity", "Other", "Inbound", True, False),
        ("PredictActionPropensity", "Multi-channel", "Multi-channel", True, True),
        (
            "PredictTreatmentPropensity",
            "Multi-channel",
            "Multi-channel",
            True,
            True,
        ),  # no longer a prediction in 24.1 release
    ]

    def getPredictionsChannelMapping(self, custom_predictions=[]):
        all_mappings = self.standardNBADPredictionsList + custom_predictions
        return pl.DataFrame(
            {
                "Prediction": [x[0].upper() for x in all_mappings],
                "Channel": [x[1] for x in all_mappings],
                "Direction": [x[2] for x in all_mappings],
                "isStandardNBADPrediction": [x[3] for x in all_mappings],
                "isMultiChannelPrediction": [x[4] for x in all_mappings],
            }
        )

    def __init__(self, df: pl.LazyFrame):
        self.predictions = (
            df.group_by(
                [
                    "Channel",
                    "ModelName",
                ]
            )
            .agg(
                pl.len(),
                Class=pl.col("ModelClass").unique(),
                Positives=pl.col("Positives")
                .filter(pl.col("ModelType") == "Prediction")
                .sum(),  # Test and Control should add up to this but they do not always do that
                Negatives=pl.col("Negatives")
                .filter(pl.col("ModelType") == "Prediction")
                .sum(),
                Positives_Control=pl.col("Positives")
                .filter(pl.col("ModelType") == "Prediction_Control")
                .sum(),
                Negatives_Control=pl.col("Negatives")
                .filter(pl.col("ModelType") == "Prediction_Control")
                .sum(),
                Positives_Test=pl.col("Positives")
                .filter(pl.col("ModelType") == "Prediction_Test")
                .sum(),
                Negatives_Test=pl.col("Negatives")
                .filter(pl.col("ModelType") == "Prediction_Test")
                .sum(),
                ResponseCount=pl.col("ResponseCount")
                .filter(pl.col("ModelType") == "Prediction")
                .sum(),
                Performance=(pl.col("Performance") * pl.col("ResponseCount"))
                .filter(pl.col("ModelType") == "Prediction")
                .sum()
                / (
                    pl.col("ResponseCount")
                    .filter(pl.col("ModelType") == "Prediction")
                    .sum()
                )
                * 100,
            )
            .with_columns(
                Class=pl.col("Class")
                .list.join(", ")
                .str.to_uppercase(),  # Prediction entries are uppercase so lets align
                CTR=pl.col("Positives") / (pl.col("Positives") + pl.col("Negatives")),
                CTR_Test=pl.col("Positives_Test")
                / (pl.col("Positives_Test") + pl.col("Negatives_Test")),
                CTR_Control=pl.col("Positives_Control")
                / (pl.col("Positives_Control") + pl.col("Negatives_Control")),
            )
            .with_columns(
                CTR_Lift=(pl.col("CTR_Test") - pl.col("CTR_Control"))
                / pl.col("CTR_Control"),
                isValidPrediction=self.validityExpr,
            )
            .sort(["ModelName"])
        )

    @property
    def is_available(self) -> bool:
        return len(self.predictions.head(1).collect()) > 0

    @property
    def is_valid(self) -> bool:
        return (
            self.is_available
            # or even stronger: pos = pos_test + pos_control
            and self.predictions.select(self.validityExpr.all()).collect().item()
        )

    def summary_by_channel(
        self,
        custom_predictions:List[NBAD_predictions]=None,
    ) -> pl.LazyFrame:
        if not custom_predictions:
            custom_predictions = []
        
        return (
            self.predictions.drop("Channel")
            .join(
                self.getPredictionsChannelMapping(custom_predictions).lazy(),
                left_on="ModelName",
                right_on="Prediction",
                how="left",
            )
            .with_columns(
                Channel=pl.when(pl.col("Channel").is_null())
                .then(pl.lit("Unknown"))
                .otherwise(pl.col("Channel")),
                Direction=pl.when(pl.col("Direction").is_null())
                .then(pl.lit("Unknown"))
                .otherwise(pl.col("Direction")),
                isStandardNBADPrediction=pl.when(
                    pl.col("isStandardNBADPrediction").is_null()
                )
                .then(pl.lit(False))
                .otherwise(pl.col("isStandardNBADPrediction")),
                isMultiChannelPrediction=pl.when(
                    pl.col("isMultiChannelPrediction").is_null()
                )
                .then(pl.lit(False))
                .otherwise(pl.col("isMultiChannelPrediction")),
            )
            .group_by(
                [
                    "ModelName",
                    "Channel",
                    "Direction",
                    "isStandardNBADPrediction",
                    "isMultiChannelPrediction",
                    "isValidPrediction",
                ]
            )
            .agg(
                cdh_utils.weighted_average_polars("CTR_Lift", "ResponseCount").alias(
                    "Lift"
                ),
                cdh_utils.weighted_performance_polars().alias("Performance"),
                pl.col("Positives").sum(),
                pl.col("Negatives").sum(),
                pl.col("ResponseCount").sum(),
                pl.col("Positives_Test").sum(),
                pl.col("Positives_Control").sum(),
                pl.col("Negatives_Test").sum(),
                pl.col("Negatives_Control").sum(),
            )
            .with_columns(CTR=(pl.col("Positives")) / (pl.col("ResponseCount")))
            .sort(["ModelName"])
        )

    def overall_summary(self, custom_predictions:List[NBAD_predictions]=None) -> pl.LazyFrame:
        return (
            self.summary_by_channel(custom_predictions)
            .filter(
                pl.col("isMultiChannelPrediction").not_() & pl.col("isValidPrediction")
            )
            .select(
                pl.concat_str(["Channel", "Direction"], separator="/")
                .n_unique()
                .alias("Number of Valid Channels"),
                cdh_utils.weighted_average_polars("Lift", "ResponseCount").alias(
                    "Overall Lift"
                ),
                cdh_utils.weighted_performance_polars().alias("Performance"),
                pl.col("Positives").sum(),
                pl.col("ResponseCount").sum(),
                pl.col("Channel")
                .where((pl.col("Lift") == pl.col("Lift").min()) & (pl.col("Lift") < 0))
                .first()
                .alias("Channel with Minimum Negative Lift"),
                pl.col("Lift")
                .where((pl.col("Lift") == pl.col("Lift").min()) & (pl.col("Lift") < 0))
                .first()
                .alias("Minimum Negative Lift"),
            )
            .with_columns(CTR=(pl.col("Positives")) / (pl.col("ResponseCount")))
        )

==> pdstools-3.4.4.tar.gz: pdstools-3.4.4/python/pdstools/prediction/Prediction.py <==
from typing import List, Optional
import polars as pl

from ..utils import cdh_utils
from ..utils import NBAD


class Prediction:
    predictions: pl.LazyFrame

    # These are pretty strict conditions - many configurations appear not to satisfy these
    # perhaps the Total = Test + Control is no longer met when Impact Analyzer is around
    prediction_validity_expr = (
        (pl.col("Positives") > 0)
        & (pl.col("Positives_Test") > 0)
        & (pl.col("Positives_Control") > 0)
        & (pl.col("Negatives") > 0)
        & (pl.col("Negatives_Test") > 0)
        & (pl.col("Negatives_Control") > 0)
        # & (
        #     pl.col("Positives")
        #     == (pl.col("Positives_Test") + pl.col("Positives_Control"))
        # )
        # & (
        #     pl.col("Negatives")
        #     == (pl.col("Negatives_Test") + pl.col("Negatives_Control"))
        # )
    )

    def getPredictionsChannelMapping(
        self, custom_predictions: Optional[List[NBAD.NBAD_Prediction]] = None
    ) -> pl.DataFrame:
        if not custom_predictions:
            custom_predictions = []
        all_mappings = NBAD.standardNBADPredictions + custom_predictions
        return pl.DataFrame(
            {
                "Prediction": [x.model_name.upper() for x in all_mappings],
                "Channel": [x.channel for x in all_mappings],
                "Direction": [x.direction for x in all_mappings],
                "isStandardNBADPrediction": [x.standard for x in all_mappings],
                "isMultiChannelPrediction": [x.multi_channel for x in all_mappings],
            }
        )

    def __init__(self, df: pl.LazyFrame):
        self.predictions = (
            df.group_by(["Channel", "ModelName", "SnapshotTime"])
            .agg(
                pl.len(),
                Class=pl.col("ModelClass").unique(),
                Positives=pl.col("Positives")
                .filter(pl.col("ModelType") == "Prediction")
                .sum(),  # Test and Control should add up to this but they do not always do that
                Negatives=pl.col("Negatives")
                .filter(pl.col("ModelType") == "Prediction")
                .sum(),
                Positives_Control=pl.col("Positives")
                .filter(pl.col("ModelType") == "Prediction_Control")
                .sum(),
                Negatives_Control=pl.col("Negatives")
                .filter(pl.col("ModelType") == "Prediction_Control")
                .sum(),
                Positives_Test=pl.col("Positives")
                .filter(pl.col("ModelType") == "Prediction_Test")
                .sum(),
                Negatives_Test=pl.col("Negatives")
                .filter(pl.col("ModelType") == "Prediction_Test")
                .sum(),
                ResponseCount=pl.col("ResponseCount")
                .filter(pl.col("ModelType") == "Prediction")
                .sum(),
                Performance=(pl.col("Performance") * pl.col("ResponseCount"))
                .filter(pl.col("ModelType") == "Prediction")
                .sum()
                / (
                    pl.col("ResponseCount")
                    .filter(pl.col("ModelType") == "Prediction")
                    .sum()
                )
                * 100,
                usesImpactAnalyzer=pl.col("ModelType").str.ends_with("_NBA").any(),
            )
            .with_columns(
                Class=pl.col("Class")
                .list.join(", ")
                .str.to_uppercase(),  # Prediction entries are uppercase so lets align
                CTR=pl.col("Positives") / (pl.col("Positives") + pl.col("Negatives")),
                CTR_Test=pl.col("Positives_Test")
                / (pl.col("Positives_Test") + pl.col("Negatives_Test")),
                CTR_Control=pl.col("Positives_Control")
                / (pl.col("Positives_Control") + pl.col("Negatives_Control")),
            )
            .with_columns(
                CTR_Lift=(pl.col("CTR_Test") - pl.col("CTR_Control"))
                / pl.col("CTR_Control"),
            )
            .sort(["ModelName"])
        )

    @property
    def is_available(self) -> bool:
        return len(self.predictions.head(1).collect()) > 0

    @property
    def is_valid(self) -> bool:
        return (
            self.is_available
            # or even stronger: pos = pos_test + pos_control
            and self.predictions.select(self.prediction_validity_expr.all())
            .collect()
            .item()
        )

    # TODO give this an optional argument for grouping by

    def summary_by_channel(
        self,
        custom_predictions: Optional[List[NBAD.NBAD_Prediction]] = None,
        keep_trend_data: bool = False,
    ) -> pl.LazyFrame:
        if not custom_predictions:
            custom_predictions = []

        return (
            self.predictions.drop("Channel")
            .join(
                self.getPredictionsChannelMapping(custom_predictions).lazy(),
                left_on="ModelName",
                right_on="Prediction",
                how="left",
            )
            .with_columns(
                Channel=pl.when(pl.col("Channel").is_null())
                .then(pl.lit("Unknown"))
                .otherwise(pl.col("Channel")),
                Direction=pl.when(pl.col("Direction").is_null())
                .then(pl.lit("Unknown"))
                .otherwise(pl.col("Direction")),
                isStandardNBADPrediction=pl.when(
                    pl.col("isStandardNBADPrediction").is_null()
                )
                .then(pl.lit(False))
                .otherwise(pl.col("isStandardNBADPrediction")),
                isMultiChannelPrediction=pl.when(
                    pl.col("isMultiChannelPrediction").is_null()
                )
                .then(pl.lit(False))
                .otherwise(pl.col("isMultiChannelPrediction")),
            )
            .group_by(
                [
                    "ModelName",
                    "Channel",
                    "Direction",
                    "isStandardNBADPrediction",
                    "isMultiChannelPrediction",
                ]
                + (["SnapshotTime"] if keep_trend_data else [])
            )
            .agg(
                cdh_utils.weighted_average_polars("CTR_Lift", "ResponseCount").alias(
                    "Lift"
                ),
                cdh_utils.weighted_performance_polars().alias("Performance"),
                pl.col("Positives").sum(),
                pl.col("Negatives").sum(),
                pl.col("ResponseCount").sum(),
                pl.col("Positives_Test").sum(),
                pl.col("Positives_Control").sum(),
                pl.col("Negatives_Test").sum(),
                pl.col("Negatives_Control").sum(),
                pl.col("usesImpactAnalyzer").any(),
            )
            .with_columns(
                ControlPercentage=100.0
                * (pl.col("Positives_Control") + pl.col("Negatives_Control"))
                / (
                    pl.col("Positives_Test")
                    + pl.col("Negatives_Test")
                    + pl.col("Positives_Control")
                    + pl.col("Negatives_Control")
                ),
                CTR=(pl.col("Positives")) / (pl.col("ResponseCount")),
                isValidPrediction=self.prediction_validity_expr,
            )
            .sort(["ModelName"] + (["SnapshotTime"] if keep_trend_data else []))
        )

    def overall_summary(
        self, custom_predictions: Optional[List[NBAD.NBAD_Prediction]] = None
    ) -> pl.LazyFrame:
        return (
            self.summary_by_channel(custom_predictions)
            .filter(
                pl.col("isMultiChannelPrediction").not_() & pl.col("isValidPrediction")
            )
            .select(
                pl.concat_str(["Channel", "Direction"], separator="/")
                .n_unique()
                .alias("Number of Valid Channels"),
                cdh_utils.weighted_average_polars("Lift", "ResponseCount").alias(
                    "Overall Lift"
                ),
                cdh_utils.weighted_performance_polars().alias("Performance"),
                pl.col("Positives").sum(),
                pl.col("ResponseCount").sum(),
                pl.col("Channel")
                .where((pl.col("Lift") == pl.col("Lift").min()) & (pl.col("Lift") < 0))
                .first()
                .alias("Channel with Minimum Negative Lift"),
                pl.col("Lift")
                .where((pl.col("Lift") == pl.col("Lift").min()) & (pl.col("Lift") < 0))
                .first()
                .alias("Minimum Negative Lift"),
                pl.col("usesImpactAnalyzer").any(),
                cdh_utils.weighted_average_polars(
                    "ControlPercentage", "ResponseCount"
                ).alias("ControlPercentage"),
            )
            .with_columns(CTR=(pl.col("Positives")) / (pl.col("ResponseCount")))
        )
